import pandas as pd

from intake_dal.dal_catalog import DalCatalog


"""
Test setup:
    - batch storage mode driver is parquet and '{{ CATALOG_DIR }}/data/user_events.parquet' has ONLY 1 row
    - local storage mode driver is csv and '{{ CATALOG_DIR }}/data/user_events.csv' has TWO rows
"""


def test_dal_catalog_default_storage_parameter(cat):
    # cat default is local -> csv plugin
    assert cat.entity.user.user_events().read().head().shape[0] == 2
    assert cat.entity.user.user_events(storage_mode="local").read().head().shape[0] == 2
    assert cat.entity.user.user_events(storage_mode="batch").read().head().shape[0] == 1
    assert cat.entity.user.user_events(storage_mode="in_mem").read().head().shape[0] == 4
    assert (
        cat.entity.user.user_events(storage_mode="local_test", data_path="data").read().head().shape[0] == 2
    )

    df = pd.DataFrame({"key": ["a", "first"], "value": [3, 42]})
    cat.entity.user.user_events(storage_mode="in_mem").write(df)
    assert cat.entity.user.user_events(storage_mode="in_mem", key="a").read().iloc[0].key == "a"
    assert cat.entity.user.user_events(storage_mode="in_mem", key="a").read().iloc[0].value == 3

    # read whole partition
    assert cat.entity.user.user_events(storage_mode="in_mem").read().shape[0] == 5


def test_dal_catalog_set_storage(catalog_path):
    cat = DalCatalog(catalog_path, storage_mode="batch")

    # ensure it's reading the batch storage
    assert cat.entity.user.user_events().read().head().shape[0] == 1

    # test storage storage_mode override:
    assert cat.entity.user.user_events(storage_mode="batch").read().head().shape[0] == 1
    assert cat.entity.user.user_events(storage_mode="local").read().head().shape[0] == 2

    # read by key
    assert cat.entity.user.user_events(storage_mode="in_mem", key="second").read().shape[0] == 1
    assert cat.entity.user.user_events(storage_mode="in_mem", key="second").read().iloc[0].key == "second"
    assert cat.entity.user.user_events(storage_mode="in_mem", key="second").read().iloc[0].value == 2

    batch_info = cat.entity.user.user_events(storage_mode="batch").discover()["metadata"]
    local_info = cat.entity.user.user_events(storage_mode="local").discover()["metadata"]

    assert batch_info["avro_schema"] == local_info["avro_schema"]
    assert batch_info["dtypes"] == local_info["dtypes"]
    assert batch_info["canonical_name"] == local_info["canonical_name"]
    assert batch_info["storage_mode"] == "batch"
    assert local_info["storage_mode"] == "local"


def test_dal_online_key_value_on_read(serving_cat: DalCatalog):
    serving_cat.entity.user.user_events(storage_mode="in_mem", key="first").read()
    pass


def test_construct_dataset(cat):
    def validate_dataset(ds):
        assert ds.name == "user_events"
        assert ds.canonical_name == "entity.user.user_events"
        assert len(ds.read()) > 0
        assert ds.cat.cat.cat == cat

    validate_dataset(cat["entity.user.user_events"])
    validate_dataset(cat.entity["user.user_events"])
    validate_dataset(cat.entity.user["user_events"])



from urllib.parse import urlparse

from intake_dal.dal_source import DalSource


def test_dal_source_description(cat):
    assert cat.entity.user.user_events.description == "user_events description"


def test_dtype(serving_cat):
    ds = serving_cat.entity.user.user_events(key="a")
    info = ds.discover()

    assert info["dtype"] == {
        "action": "object",
        "home_id": "int32",
        "timestamp": "datetime64",
        "userid": "int64",
    }


def test_avro_schema(serving_cat):
    def validate_avro_schema(schema):
        assert "fields" in schema
        assert schema["name"] == "Root"
        assert schema["type"] == "record"
        assert schema["fields"][0]["name"] == "userid"
        assert schema["fields"][0]["type"] == "long"
        assert schema["fields"][1]["name"] == "home_id"
        assert schema["fields"][1]["type"] == "int"

    ds = serving_cat.entity.user.user_events(key="a")
    info = ds.discover()
    validate_avro_schema(info["metadata"]["avro_schema"])
    validate_avro_schema(ds.avro_schema)


def test_canonical_name(serving_cat):
    ds = serving_cat.entity.user.user_events(key="a")
    assert ds.discover()["metadata"]["canonical_name"] == "entity.user.user_events"


def test_parse_storage_mode_url():
    def validate_url(url: str, expected: str):
        assert DalSource.parse_storage_mode_url(url) == (urlparse(url), expected)

    validate_url("csv://{{ CATALOG_DIR }}/data/user_events.csv", "{{ CATALOG_DIR }}/data/user_events.csv")

    validate_url("dal-online://http://0.0.0.0:9166#zpid", "http://0.0.0.0:9166#zpid")

    validate_url(
        "hive://user_events_dal_catalog2;userid={{userid}}?q1=v1#fragment",
        "user_events_dal_catalog2;userid={{userid}}?q1=v1#fragment",
    )



import pandas as pd
import pkg_resources
from intake import DataSource, Schema


class InMemoryKVSource(DataSource):
    container = "dataframe"
    version = pkg_resources.get_distribution("intake-dal").version
    partition_access = False
    name = "in-memory-kvs"

    db = pd.DataFrame({"key": ["first", "second", "third", "fourth"], "value": [1, 2, 3, 4]})

    def __init__(self, urlpath="", key=None, storage_options=None, metadata=None):
        # store important kwargs
        self._urlpath = urlpath
        self._key = key
        super(InMemoryKVSource, self).__init__(metadata=metadata)

    def _get_schema(self):
        self._dtypes = {"key": "object", "value": "object"}

        return Schema(
            datashape=None,
            dtype=self._dtypes,
            shape=(None, len(self._dtypes)),
            npartitions=1,  # This data is not partitioned, so there is only one partition
            extra_metadata={},
        )

    def _get_partition(self, _) -> pd.DataFrame:
        if self._key:
            return InMemoryKVSource.db[self.db.key == self._key]
        else:
            return InMemoryKVSource.db

    def write(self, df: pd.DataFrame):
        new_db = pd.merge(InMemoryKVSource.db, df, on="key", how="outer")
        new_db["value"] = new_db["value_y"].fillna(new_db["value_x"]).astype("int")

        InMemoryKVSource.db = new_db[["key", "value"]]
        return InMemoryKVSource.db

    def _close(self):
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['intake_dal']

package_data = \
{'': ['*'], 'intake_dal': ['tests/*', 'tests/data/*']}

install_requires = \
['deepmerge==0.1.0',
 'intake-nested-yaml-catalog>=0.1.0',
 'intake>=0.5.3',
 'pandavro>=1.5.1,<2.0.0',
 'pyarrow>=0.15.1,<0.16.0',
 'pyyaml==5.1.2',
 'toolz>=0.10.0,<0.11.0',
 'vcver>=0.2.10']

entry_points = \
{'intake.drivers': ['dal = intake_dal.dal_source:DalSource',
                    'dal-online = intake_dal.dal_online:DalOnlineSource',
                    'dal_cat = intake_dal.dal_catalog:DalCatalog',
                    'in-memory-kvs = intake_dal.in_memory_kv:InMemoryKVSource']}

setup_kwargs = {
    'name': 'intake-dal',
    'version': '0.1.4',
    'description': 'Intake single YAML hierarchical catalog.',
    'long_description': '.. image:: https://travis-ci.org/zillow/intake-dal.svg?branch=master\n    :target: https://travis-ci.org/zillow/intake-dal\n\n.. image:: https://coveralls.io/repos/github/zillow/intake-dal/badge.svg?branch=master\n    :target: https://coveralls.io/github/zillow/intake-dal?branch=master\n\n.. image:: https://readthedocs.org/projects/intake-dal/badge/?version=latest\n    :target: https://intake-dal.readthedocs.io/en/latest/?badge=latest\n    :alt: Documentation Status\n\n\nWelcome to Intake DAL (data access layer) plugin\n==================================================\nThis `Intake <https://intake.readthedocs.io/en/latest/quickstart.html>`_ plugin helps\nabstract a dataset over disparate storage systems (eg: bulk, streaming, serving, ...).\nIt also provides an easy way to specialize a\n`hierarchical catalog <https://github.com/zillow/intake-nested-yaml-catalog/>`_\nto a default DAL storage system.\n\n\nSample Catalog source entry:\n\n.. code-block:: yaml\n\n    user_events:\n      driver: dal\n      args:\n        default: \'local\'\n        storage:\n          local: \'csv://{{ CATALOG_DIR }}/data/user_events.csv\'\n          serving: \'in-memory-kv://foo\'\n          batch: \'parquet://{{ CATALOG_DIR }}/data/user_events.parquet\'\n\nExample code using sample catalog:\n\n.. code-block:: python\n\n  # Specialize the catalog dal default storge mode datasources\n  # to be "serving".\n  cat = DalCatalog(path, storage_mode="serving")\n\n  # reads from the serving storage system\n  # using the in-memory-kv Intake plugin\n  df = cat.user_events.read()\n\n\n',
    'author': 'Zillow AI Platform',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6',
}


setup(**setup_kwargs)



from pathlib import Path

import pytest

from intake_dal.dal_catalog import DalCatalog


@pytest.fixture
def catalog_path():
    return str(Path(__file__).resolve().parent.joinpath(Path("catalog.yaml")))


@pytest.fixture
def serving_cat(catalog_path: str):
    return DalCatalog(catalog_path, storage_mode="serving")


@pytest.fixture
def cat(catalog_path: str):
    return DalCatalog(catalog_path)
